@@ -10,7 +10,11 @@ import string
 
 from error import IIIFError
 
-class IIIFRequest:
+class IIIFRequestBaseURI(Exception):
+    """ Subclass of Exception to indicate request for base URI """
+    pass
+
+class IIIFRequest(object):
     """ Implement IIIF request URL syntax
     
     There are two URL forms defined in section 2:
@@ -18,32 +22,37 @@ class IIIFRequest:
     or
     http[s]://server/[prefix/]identifier/info.format
 
+    We also detect and throw a special IIIFRequestBaseURI exception then
+    the naked base URI is requested:
+    http[s]://server/[prefix/]identifier(/?)
+
     The attribites of objects of this class follow the names here except that
     baseurl is used for "http[s]://server/[prefix/]". If baseurl is not set 
     then a relative URL will be created or parsed.
     """
 
-    def __init__(self, **params):
+    def __init__(self, api_version='2.0', **params):
         """Create Request object and optionally set any attributes via 
-        named parameters
+        named parameters.
 
+        Current API version assumed ('2.0') if not specified. If another API
+        version is to be used then this should be set on creation via the
+        api_version parameter.
+ 
         Unless specified the baseurl will be set to nothing ("").
         """
-        self.clear()
-        self.region_full=False
-        self.region_pct=False
-        self.region_xywh=None # (x,y,w,h)
-        self.size_wh=None     # (w,h)
-        #
         self.baseurl = ''
+        self.api_version = api_version
+        self.clear()
         self.set(**params)
 
     def clear(self):
-        """ Clear all data that might pertain to an individual iiif URL
+        """ Clear all data that might pertain to an individual IIIF URL
 
-        Does not change/reset the baseurl which might be useful in a
-        sequence of calls.
+        Does not change/reset the baseurl or API version which might be 
+        useful in a sequence of calls.
         """
+        # API parameters
         self.identifier = None
         self.region = None
         self.size = None
@@ -51,15 +60,41 @@ class IIIFRequest:
         self.quality = None
         self.format = None
         self.info = None
-        # Other flags
+        # Derived data and flags
+        self.region_full = False
+        self.region_pct = False
+        self.region_xywh = None # (x,y,w,h)
         self.size_full = False
         self.size_pct = None
         self.size_bang = None
+        self.size_wh = None     # (w,h)
+        self.rotation_mirror = False
         self.rotation_deg = 0.0 
 
+    @property
+    def api_version(self):
+        """ Get api_version value """
+        return self._api_version
+
+    @api_version.setter
+    def api_version(self,v):
+        """ Set the api_version and associated configurations """
+        self._api_version=v
+        if (self._api_version>='2.0'):
+            self.default_quality = 'default'
+            self.allowed_qualities = ['default','color','bitonal','gray']
+        else: # versions 1.0 and 1.1
+            self.default_quality = 'native'
+            self.allowed_qualities = ['native','color','bitonal','grey']
+
     def set(self, **params):
-        # FIXME - maybe make this safe an allow only setting valid attributes
-        self.__dict__.update(**params)
+        """ Set one of the allowed request parameters
+
+        Will silently ignore any unknown parameters.
+        """
+        for k in ('identifier','region','size','rotation','quality','format','info'):
+            if (k in params):
+                setattr(self, k, params[k])
 
     def quote(self, path_segment):
         """ Quote parameters in IIIF URLs
@@ -92,7 +127,7 @@ class IIIFRequest:
             if self.region:
                 region = self.region
             elif self.region_xywh:
-                region = "%d,%d,%d,%d" % (self.region_xywh[0],self.region_xywh[1],self.region_xywh[2],self.region_xywh[3]) #FIXME - how to get tiple from list
+                region = "%d,%d,%d,%d" % tuple(self.region_xywh)
             else:
                 region = "full"
             if self.size:
@@ -102,7 +137,7 @@ class IIIFRequest:
             else:
                 size = "full" 
             rotation = self.rotation if self.rotation else "0"
-            quality = self.quality if self.quality else "native"
+            quality = self.quality if self.quality else self.default_quality
             # parameterized form
             path += self.quote(region) + "/" +\
                     self.quote(size) + "/" +\
@@ -141,36 +176,51 @@ class IIIFRequest:
             if (num != 1):
                 raise(IIIFError("URL does not match baseurl (server/prefix)."))
             url=path
-        # Look for optional .fmt at end, must start with letter. Note that we 
-        # want to catch the case of a dot and no format (format='') which is
-        # different from no dot (format=None)
-        m = re.match( "(.+)\.([a-zA-Z]\w*)$", url)
-        if (m):
-            # There is a format string at end, chop off and store
-            url = m.group(1)
-            self.format = ( m.group(2) if (m.group(2) is not None) else '')
         # Break up by path segments, count to decide format
         segs = string.split( url, '/', 5)
         if (len(segs) > 5):
-            raise(IIIFError(code=400,text="Too many path segments in URL (got %d: %s) in URL."%(len(segs),' | '.join(segs))))
+            raise(IIIFError(code=404,text="Too many path segments in URL (got %d: %s) in URL."%(len(segs),' | '.join(segs))))
         elif (len(segs) == 5):
             self.identifier = urllib.unquote(segs[0])
             self.region = urllib.unquote(segs[1])
             self.size  = urllib.unquote(segs[2])
             self.rotation = urllib.unquote(segs[3])
-            self.quality = urllib.unquote(segs[4])
+            self.quality = self.strip_format(urllib.unquote(segs[4]))
             self.info = False
         elif (len(segs) == 2):
             self.identifier = urllib.unquote(segs[0])
-            if (urllib.unquote(segs[1]) != "info"):
+            info_name = self.strip_format(urllib.unquote(segs[1]))
+            if (info_name != "info"):
                 raise(IIIFError(code=400,text="Badly formed information request, must be info.json or info.xml"))
-            if (self.format not in ("json","xml")):
-                raise(IIIFError(code=400,text="Bad information request format, must be json or xml"))
+            if (self.api_version=='1.0'):
+                if (self.format not in ['json','xml']):
+                    raise(IIIFError(code=400,text="Bad information request format, must be json or xml"))
+            elif (self.format!='json'):
+                raise(IIIFError(code=400,text="Bad information request format, must be json"))
             self.info = True
+        elif (len(segs) == 1):
+            self.identifier = urllib.unquote(segs[0])
+            raise(IIIFRequestBaseURI())
         else:
             raise(IIIFError(code=400,text="Bad number of path segments (%d: %s) in URL."%(len(segs),' | '.join(segs))))
         return(self)
 
+    def strip_format(self,str_and_format):
+        """ Look for optional .fmt at end
+
+        The format must start with letter. Note that we want to catch 
+        the case of a dot and no format (format='') which is different 
+        from no dot (format=None)
+        
+        Sets self.format as side effect, returns possibly modified string
+        """
+        m = re.match( "(.+)\.([a-zA-Z]\w*)$", str_and_format)
+        if (m):
+            # There is a format string at end, chop off and store
+            str_and_format = m.group(1)
+            self.format = ( m.group(2) if (m.group(2) is not None) else '')
+        return(str_and_format)
+
     def parse_parameters(self):
         """ Parse the parameters of a parameterized request
 
@@ -209,7 +259,7 @@ class IIIFRequest:
         str_values = string.split(xywh, ',', 5)
         if (len(str_values) != 4):
             raise IIIFError(code=400,parameter="region",
-                           text="Bad number of values in region specification, must be x,y,w,h but got %d value(s) from '%s'"%(len(str_values),xywh))
+                            text="Bad number of values in region specification, must be x,y,w,h but got %d value(s) from '%s'"%(len(str_values),xywh))
         values=[]
         for str_value in str_values:
             # Must be either integer (not pct) or interger/float (pct)
@@ -222,24 +272,24 @@ class IIIFRequest:
                                    text="Bad floating point value for percentage in region (%s)."%str_value)
                 if (value>100.0):
                     raise IIIFError(code=400,parameter="region",
-                                   text="Percentage over value over 100.0 in region (%s)."%str_value)
+                                    text="Percentage over value over 100.0 in region (%s)."%str_value)
             else:
                 try:
                     value=int(str_value)
                 except ValueError:
                     raise IIIFError(code=400,parameter="region",
-                                   text="Bad integer value in region (%s)."%str_value)
+                                    text="Bad integer value in region (%s)."%str_value)
             if (value<0):
                 raise IIIFError(code=400,parameter="region",
-                               text="Negative values not allowed in region (%s)."%str_value)
+                                text="Negative values not allowed in region (%s)."%str_value)
             values.append(value)
         # Zero size region is w or h are zero (careful that they may be float)
         if (values[2]==0.0 or values[3]==0.0):
             raise IIIFError(code=400,parameter="region",
-                           text="Zero size region specified (%s))."%xywh)
+                            text="Zero size region specified (%s))."%xywh)
         self.region_xywh=values
 
-    def parse_size(self):
+    def parse_size(self,size=None):
         """Parse the size component of the path
 
         /full/ -> self.size_full = True
@@ -257,6 +307,8 @@ class IIIFRequest:
               # scale to w by h
         Returns (None,None) if no scaling is required.
         """
+        if (size is not None):
+            self.size = size
         self.size_pct=None
         self.size_bang=False
         self.size_full=False
@@ -271,22 +323,23 @@ class IIIFRequest:
                 self.size_pct=float(pct_str)
             except ValueError:
                 raise IIIFError(code=400,parameter="size",
-                               text="Percentage size value must be a number, got '%s'."%(pct_str))
+                                text="Percentage size value must be a number, got '%s'."%(pct_str))
 # FIXME - current spec places no upper limit on size
 #            if (self.size_pct<0.0 or self.size_pct>100.0):
 #                raise IIIFError(code=400,parameter="size",
 #                               text="Illegal percentage size, must be 0 <= pct <= 100.")
             if (self.size_pct<0.0):
                 raise IIIFError(code=400,parameter="size",
-                               text="Base size percentage, must be > 0.0, got %f."%(self.size_pct))
+                                text="Base size percentage, must be > 0.0, got %f."%(self.size_pct))
         else:
-            bang_match = re.match('!(.*)$',self.size)
-            if (bang_match is not None):
+            if (self.size[0]=='!'):
                 # Have "!w,h" form
-                (mw,mh)=self._parse_w_comma_h(bang_match.group(1),'size')
+                size_no_bang=self.size[1:]
+                (mw,mh)=self._parse_w_comma_h(size_no_bang,'size')
                 if (mw is None or mh is None):
                     raise IIIFError(code=400,parameter="size",
-                                   text="Illegal size requested: both w,h must be specified in !w,h requests.")
+                                    text="Illegal size requested: both w,h must be specified in !w,h requests.")
+                self.size_wh=(mw,mh)
                 self.size_bang=True
             else:
                 # Must now be "w,h", "w," or ",h"
@@ -296,25 +349,24 @@ class IIIFRequest:
             if ( ( w is not None and w<=0) or 
                  ( h is not None and h<=0) ):
                 raise IIIFError(code=400,parameter='size',
-                               text="Size parameters request zero size result image.")
+                                text="Size parameters request zero size result image.")
 
-    def _parse_w_comma_h(self,str,param):
+    def _parse_w_comma_h(self,whstr,param):
         """ Utility to parse "w,h" "w," or ",h" values
         
         Returns (w,h) where w,h are either None or ineteger. Will
         throw a ValueError if there is a problem with one or both.
         """
-        (wstr,hstr) = string.split(str, ',', 2)
         try:
+            (wstr,hstr) = string.split(whstr, ',', 2)
             w = self._parse_non_negative_int(wstr,'w')
             h = self._parse_non_negative_int(hstr,'h')
         except ValueError as e:
             raise IIIFError(code=400,parameter=param,
-                           text="Illegal parameter value (%s)." % str(e) )
+                            text="Illegal %s value (%s)." % (param,str(e)) )
         if (w is None and h is None):
             raise IIIFError(code=400,parameter=param,
-                           text="Must specify at least one of w,h.")
-        self.size_wh=(w,h)
+                            text="Must specify at least one of w,h for %s." % (param))
         return(w,h)
 
     def _parse_non_negative_int(self,istr,name):
@@ -333,23 +385,34 @@ class IIIFRequest:
             raise ValueError("Illegal negative value for %s" % (name))
         return(i)
 
-    def parse_rotation(self):
+    def parse_rotation(self, rotation=None):
         """ Check and interpret rotation
 
-        Sets self.rotation_deg to a floating point number 0 <= angle < 360. Includes
-        translation of 360 to 0.
+        Uses value of self.rotation at starting point unless rotation parameter
+        is specified in the call. Sets self.rotation_deg to a floating point 
+        number 0 <= angle < 360. Includes translation of 360 to 0. If there is 
+        a prefix bang (!) then self.rotation_mirror will be set True, otherwise
+        it will be False.
         """
+        if (rotation is not None):
+            self.rotation = rotation
+        self.rotation_deg=0.0
+        self.rotation_mirror=False
         if (self.rotation is None):
-            self.rotation_deg=0.0
             return
+        # Look for ! prefix first
+        if (self.rotation[0]=='!'):
+            self.rotation_mirror=True
+            self.rotation=self.rotation[1:]
+        # Interpret value now
         try:
             self.rotation_deg=float(self.rotation)
         except ValueError:
             raise IIIFError(code=400,parameter="rotation",
-                           text="Bad rotation value, must be a number, got '%s'."%(self.rotation))
+                            text="Bad rotation value, must be a number, got '%s'."%(self.rotation))
         if (self.rotation_deg<0.0 or self.rotation_deg>360.0):
             raise IIIFError(code=400,parameter="rotation",
-                           text="Illegal rotation value, must be 0 <= rotation <= 360, got %f."%(self.rotation_deg))
+                            text="Illegal rotation value, must be 0 <= rotation <= 360, got %f."%(self.rotation_deg))
         elif (self.rotation_deg==360.0):
             # The spec admits 360 as valid, but change to 0
             self.rotation_deg=0.0
@@ -361,10 +424,10 @@ class IIIFRequest:
         default. Checks for the three valid values else throws and IIIFError.
         """
         if (self.quality is None):
-            self.quality_val='native'
-        elif (self.quality not in ['native','color','bitonal','grey']):
+            self.quality_val=self.default_quality
+        elif (self.quality not in self.allowed_qualities):
             raise IIIFError(code=400,parameter="quality",
-                           text="The quality parameter must be 'native', 'color', 'bitonal' or 'grey', got '%s'."%(self.quality))
+                            text="The quality parameter must be '%s', got '%s'."%("', '".join(self.allowed_qualities),self.quality))
         else:
             self.quality_val=self.quality
 
